/**
 *
   serial.cc, com port objects
   Copyright (C)  2010    Angelo Dureghello (angelo70@gmail.com)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include "serial.hh"

#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <cstdio>
#include <cerrno>
#include <wx/config.h>

#include <fcntl.h>		/* File control definitions */
#include <termios.h>		/* POSIX terminal control definitions */

using std::string;
using std::stringstream;

/* clone fd, allows read and write */
ComPort::ComPort (ComPort &c)
{
   fd = dup(c.GetFd());
}

ComPort::~ComPort ()
{
   close (fd);
}

ComPort::ComPort (int number, ComSpeed speed, ComParity parity, int data_bits, int stop_bits)
{
	stringstream port;

#ifndef WIN32
	port << "/dev/ttyS";
#else
	port << "COM";
#endif
	port << number;

	InitPort(port.str().c_str(), speed, parity, data_bits, stop_bits);
}

ComPort::ComPort (const char *port,
						ComSpeed speed,
						ComParity parity,
						int data_bits,
						int stop_bits)
{
	InitPort(port, speed, parity, data_bits, stop_bits);
}
/**
 * This function open the port in fully raw mode, to send binary data also.
 */
void ComPort::InitPort ( const char *port,
								 ComSpeed speed,
								 ComParity parity,
								int data_bits,
								int stop_bits )
{
    /* store speed val */
    speed_val = SpeedToInt(speed);

    wxConfig::Get()->SetPath("/Terminal");

    fd = open(port, O_RDWR | O_NOCTTY | O_NDELAY);

    if (fd == -1)
    {
        stringstream msg;

        msg << "ComPort::" << __func__ << " unable to open port: " << port;

        perror(msg.str().c_str());

        return;
    }
    else
    {
        /* read return immediately (non-blocking) */
        fcntl (fd, F_SETFL, FNDELAY);

        struct termios options;

        tcgetattr(fd, &options);

        int s = TranslateSpeed(speed);

        cfsetispeed(&options, s);
        cfsetospeed(&options, s);

        options.c_cflag |= (CLOCAL | CREAD);
        options.c_cflag &= ~CSIZE;
        // no flow control
        options.c_cflag &= ~CRTSCTS;
        options.c_cflag &= ~TranslateParity(parity);

        (stop_bits>=2) ? options.c_cflag |= CSTOPB : options.c_cflag &= ~CSTOPB;

        switch (data_bits)
        {
            case 5:	options.c_cflag |= CS5;	break;
            case 6:	options.c_cflag |= CS6;	break;
            case 7:	options.c_cflag |= CS7;	break;
            case 8:
            default: options.c_cflag |= CS8;	break;
        }

        // turn off s/w flow ctrl
        options.c_iflag &= ~(IXON | IXOFF | IXANY);

        bool val;

        if (wxConfig::Get()->Read("nl_to_cr",&val)==true)
        {
            if (val)
                options.c_iflag |= INLCR;
            else
                options.c_iflag &= ~INLCR;
        }

        options.c_oflag = 0;

        options.c_cc[VMIN]=0;
        options.c_cc[VTIME]=0;

        // make raw
        options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
        options.c_oflag &= ~OPOST;

        tcflush		(fd, TCIFLUSH);
        tcsetattr	(fd, TCSANOW, &options);
    }
}

void ComPort::SendBreak ()
{
	if (fd)
		tcsendbreak(fd, 0);
}

bool ComPort::ToggleDTR ()
{
	if (fd)
	{
		int bits;

		ioctl(fd, TIOCMGET, &bits);
		(bits&TIOCM_DTR) ? bits&=~TIOCM_DTR : bits|=TIOCM_DTR ;
		ioctl(fd, TIOCMSET, &bits);
		return bits&TIOCM_DTR;
	}
	return false;
}

bool ComPort::ToggleRTS ()
{
	if (fd)
	{
		int bits;

		ioctl(fd, TIOCMGET, &bits);
		(bits&TIOCM_RTS) ? bits&=~TIOCM_RTS : bits|=TIOCM_RTS ;
		ioctl(fd, TIOCMSET, &bits);
		return bits&TIOCM_RTS;
	}
	return false;
}

int ComPort::GetModemFlags ()
{
	int bits;

	ioctl(fd, TIOCMGET, &bits);
	return bits;
}

int ComPort::TranslateSpeed (ComSpeed s)
{
#ifndef WIN32
	switch (s)
	{
		case CS200:		return B200;
		case CS300:		return B300;
		case CS600:		return B600;
		case CS1200:	return B1200;
		case CS2400:	return B2400;
		case CS4800:	return B4800;
		case CS9600:	return B9600;
		case CS19200:	return B19200;
		case CS38400:	return B38400;
		case CS57600:	return B57600;
		case CS115200:	return B115200;
		case CS230400: return B230400;
		default: return 0;
	}
#else
	/* TO DO */
#endif
}

string ComPort::GetSpeedAsString (ComSpeed s)
{
	switch (s)
	{
		case CS200:		return "200";
		case CS300:		return "300";
		case CS600:		return "600";
		case CS1200:	return "1200";
		case CS2400:	return "2400";
		case CS4800:	return "4800";
		case CS9600:	return "9600";
		case CS19200:	return "19200";
		case CS38400:	return "38400";
		case CS57600:	return "57600";
		case CS115200:	return "115200";
		case CS230400: return "230400";
		default: return "";
	}
	return "";
}

string ComPort::GetParityAsString (ComParity p)
{
   switch (p) {
      case 0: return "O";
      case 1: return "E";
      case 2: return "N";
   }
   return "";
}

int ComPort::TranslateParity	(ComParity p)
{
#ifndef WIN32
	switch (p)
	{
		case CPODD:		return PARODD;
		case CPEVEN:	return PARENB;
		case CPNONE:
		default: return 0;
	}
#else
	/* TO DO */
#endif
}

int ComPort::Write (const char *buff, int len)
{
    int w;

    do {
        w = write(fd, buff, len);
        if (w>=0) {
            len  -= w;
            buff += w;
        }
        else {
            switch (errno)
            {
            default:
            case EIO: return -1; /* serial porrt exist but not writeable */
            }
        }
    }
    while (len>0);

    return 1;
}

int ComPort::Read (char *buff, int max_size)
{
	return read(fd, buff, max_size);
}

